/* $Id: clientconnectionlistener.hpp,v 1.2 2005/06/28 13:55:18 chfreund Exp $ */

#ifndef _CLIENTCONNECTIONLISTENER_HPP_
#define _CLIENTCONNECTIONLISTENER_HPP_

#include <SDL.h>
#include <SDL_net.h>

class ServerCommunicator;
class ClientConnection;

//! Abstract base class for connection listeners
class ClientConnectionListener {
public:
	//! Destructor
	virtual ~ClientConnectionListener() { }
	//! Check for new connections
	//! \return A new connection if available, and 0 else
	virtual ClientConnection* getNewConnection( Uint8 clientID ) = 0;
};

//! Class that listens for incoming TCP connections
class ClientConnectionListenerTCP : public ClientConnectionListener {
private:
	//! The communicator responsible for this listener
	ServerCommunicator* m_communicator;
	//! The TCP port that is listened on
	Uint16 m_port;
	//! A socket for listening
	TCPsocket m_socket;
	//! A socket set
	SDLNet_SocketSet m_socketSet;
	//! \name Members for handling bait packets
	//@{
	//! A UDP socket
	UDPsocket m_baitSocket;
	//! A socket set for the UDP socket
	SDLNet_SocketSet m_baitSocketSet;
	//! Storage for bait packet
	UDPpacket* m_baitPacket;
	//@}
public:
	//! Constructor
	//! \param port The TCP port number for listening
	ClientConnectionListenerTCP( ServerCommunicator* communicator, Uint16 port );
	//! Destructor
	virtual ~ClientConnectionListenerTCP();
	ClientConnection* getNewConnection( Uint8 clientID );
private:
	//! Handle incoming bait packets
	void handleBaitPackets();
};

#endif // _CLIENTCONNECTIONLISTENER_HPP_
